from abc import ABC


class Tao(ABC):
    """the lion way"""


class AbstractSpace(Tao):
    """Abstract space or context."""


class AbstractElement(Tao):
    """Entity within a space."""

    @classmethod
    def class_name(cls) -> str:
        """
        Get class name.

        Returns:
            str: The name of the class.
        """
        return cls.__name__


# must have async abstract methods
class AbstractObserver(Tao):
    """Entity capable of making observations"""


# must have async abstract methods
class AbstractObservation(Tao):
    """Act of observing."""


__all__ = [
    "AbstractSpace",
    "AbstractElement",
    "AbstractObserver",
    "AbstractObservation",
]



from .characteristic import Real
from .concept import AbstractElement


class BaseRecord(AbstractElement, Real):
    """Base class for records."""


class MutableRecord(BaseRecord):
    """Mutable record class."""


class ImmutableRecord(BaseRecord):
    """Immutable record class."""


__all__ = [
    "BaseRecord",
    "MutableRecord",
    "ImmutableRecord",
]



from abc import abstractmethod
from collections.abc import Iterable
from typing import Any, Generic, TypeVar

from .characteristic import Traversal
from .concept import AbstractElement, AbstractSpace

T = TypeVar("T")


class Container(AbstractSpace, AbstractElement, Generic[T]):
    """Container for items."""

    @abstractmethod
    def __contains__(self, item: object) -> bool:
        """Check if an item is in the space."""


class Ordering(Container, Generic[T]):
    """Container with a defined order. Subclass must have order attribute."""

    order: list[T]


class Collective(Container, Generic[T]):
    """Container representing a collection of items."""

    @abstractmethod
    def items(self) -> Iterable[tuple[Any, T]]:
        """
        Get the items in the collective.

        Returns:
            Iterable: The items in the collective.
        """

    @abstractmethod
    def values(self) -> Iterable[T]:
        """
        Get the values in the collective.

        Returns:
            Iterable: The values in the collective.
        """

    @abstractmethod
    def keys(self) -> Iterable[Any]:
        """
        Get the keys in the collective.

        Returns:
            Iterable: The keys in the collective.
        """


class Structure(Container, Traversal, Generic[T]):
    """Traversable container structure"""


__all__ = ["Container", "Ordering", "Collective", "Structure"]



import asyncio
from abc import abstractmethod
from typing import Any

from .concept import AbstractObserver
from .observation import Event


class BaseManager(AbstractObserver):
    """Coordinates other system components."""


class BaseProcessor(AbstractObserver):
    """Observers for information transformation and analysis."""

    event_type: type[Event]

    def __init__(self, capacity: int, refresh_time: float) -> None:
        """
        Initializes an ActionProcessor instance.

        Args:
            capacity: Max number of actions processed concurrently.
            refresh_time: Time interval between processing cycles.

        Raises:
            ValueError: If capacity < 0 or refresh_time is negative.
        """
        if capacity < 0:
            raise ValueError("initial capacity must be >= 0")
        if refresh_time < 0:
            raise ValueError("refresh time for execution can not be negative")

        self.capacity = capacity
        self.queue = asyncio.Queue()
        self._stop_event = asyncio.Event()
        self.available_capacity = capacity
        self.execution_mode: bool = False
        self.refresh_time = refresh_time

    async def enqueue(self, event: Event) -> None:
        """
        Enqueues an event to the processor queue.

        Args:
            event: The event to be added to the queue.
        """
        await self.queue.put(item=event)

    async def dequeue(self) -> Event:
        """
        Dequeues an event from the processor queue.

        Returns:
            The next event in the queue.
        """
        return await self.queue.get()

    async def join(self) -> None:
        """Blocks until all items in the queue have been processed."""
        await self.queue.join()

    async def stop(self) -> None:
        """Signals the processor to stop processing actions."""
        self._stop_event.set()

    async def start(self) -> None:
        """Allows the processor to start or continue processing."""
        self._stop_event.clear()

    def is_stopped(self) -> bool:
        """
        Indicates whether the processor has been stopped.

        Returns:
            True if the processor has been stopped, otherwise False.
        """
        return self._stop_event.is_set()

    @classmethod
    async def create(cls, **kwargs: Any) -> "BaseProcessor":
        """
        Class method to create an instance of the processor.

        Args:
            **kwargs: Arguments passed to the processor constructor.

        Returns:
            A new instance of the processor.
        """
        processor = cls(**kwargs)
        return processor

    @abstractmethod
    async def process(self) -> None:
        """process the event."""
        pass

    async def request_permission(self, **kwargs: Any) -> bool:
        """
        Placeholder method to request permission before processing an event.

        Args:
            **kwargs: Arbitrary keyword arguments for requesting permission.

        Returns:
            Always returns True, indicating permission is granted.
        """
        return True

    async def execute(self) -> None:
        """
        Executes the processor, continuously processing actions until stopped.

        Runs in a loop, processing actions and respecting the refresh time
        between cycles. Exits when signaled to stop.
        """
        self.execution_mode = True
        await self.start()

        while not self.is_stopped():
            await self.process()
            await asyncio.sleep(self.refresh_time)
        self.execution_mode = False


class BaseExecutor(AbstractObserver):
    """Performing tasks with `processor`."""

    processor_class: type[BaseProcessor]
    strict: bool = True

    def __init__(self, **kwargs: Any) -> None:
        """
        Initializes the ActionExecutor with the provided configuration.

        Args:
            **kwargs: Configuration parameters for initializing the processor.
        """
        self.processor_config = kwargs

    @abstractmethod
    async def forward(self, *args: Any, **kwargs: Any) -> Any:
        """Move onto the next step."""

    async def create_processor(self) -> None:
        """Factory method the processor creation"""
        self.processor = await self.processor_class.create(
            **self.processor_config,
        )

    async def start(self) -> None:
        """Starts the event processor."""
        if not self.processor:
            await self.create_processor()
        await self.processor.start()

    async def stop(self) -> None:
        """Stops the event processor."""
        if self.processor:
            await self.processor.stop()


class BaseEngine(AbstractObserver):

    @abstractmethod
    async def run(self, *args: Any, **kwargs: Any) -> Any:
        """Asynchronously runs the engine"""


class BaseOperator(AbstractObserver):

    @abstractmethod
    async def operate(self, *args: Any, **kwargs: Any) -> Any:
        """Asynchronously operate intelligent models"""


__all__ = [
    "BaseManager",
    "BaseExecutor",
    "BaseProcessor",
    "BaseOperator",
    "BaseEngine",
]



from .concept import Tao


class Characteristic(Tao):
    """Base class for Lion framework characteristics."""


class Observable(Characteristic):
    """Identifiable objects with a 'ln_id' attribute."""

    ln_id: str


class Temporal(Characteristic):
    """objects existing in time with a 'timestamp' attribute."""

    timestamp: float


class Real(Observable, Temporal):
    """objects have definite properties independent of observation"""


class Communicatable(Real):
    """Represents entities that can be used for communication."""

    sender: str
    recipient: str


class Quantum(Characteristic):
    """quantum-inspired approaches for modeling of non-classical behaviors."""


class Probabilistic(Characteristic):
    """entities with inherent uncertainty."""


class Stochastic(Probabilistic, Temporal):
    """Combines probabilistic and time-dependent aspects."""


class Relational(Real):
    """only relational entities can be nodes in a graph"""


class Traversal(Characteristic):
    """progressable or explorable entities"""


__all__ = [
    "Observable",
    "Temporal",
    "Relational",
    "Traversal",
    "Communicatable",
    # "Quantum",
    # "Probabilistic",
    # "Stochastic",
]



"""The abstract base classes for the LION framework."""

from .characteristic import (
    Communicatable,
    Observable,
    Real,
    Relational,
    Temporal,
    Traversal,
)
from .concept import (
    AbstractElement,
    AbstractObservation,
    AbstractObserver,
    AbstractSpace,
)
from .observation import Action, Condition, Event, EventStatus, Signal
from .observer import (
    BaseEngine,
    BaseExecutor,
    BaseManager,
    BaseOperator,
    BaseProcessor,
)
from .record import BaseRecord, ImmutableRecord, MutableRecord
from .space import Collective, Container, Ordering, Structure

__all__ = [
    "AbstractElement",
    "AbstractObservation",
    "AbstractObserver",
    "AbstractSpace",
    "Action",
    "BaseEngine",
    "BaseExecutor",
    "BaseManager",
    "BaseProcessor",
    "BaseOperator",
    "Condition",
    "Container",
    "Collective",
    "Event",
    "EventStatus",
    "Observable",
    "Ordering",
    "Relational",
    "Signal",
    "Structure",
    "Temporal",
    "Traversal",
    "ImmutableRecord",
    "BaseRecord",
    "MutableRecord",
    "Communicatable",
    "Real",
]



from abc import abstractmethod
from enum import Enum
from typing import Any

from .concept import AbstractObservation


class EventStatus(str, Enum):
    PENDING = "pending"
    PROCESSING = "processing"
    COMPLETED = "completed"
    FAILED = "failed"


class Event(AbstractObservation):
    """Discrete occurrences or state changes."""

    status: EventStatus

    @property
    def request(self) -> dict:
        """Retrieve the permission request payload."""
        return self._request()

    def _request(self) -> dict:
        """override this method in child class."""
        return {}


class Condition(Event):
    """State evaluation."""

    @abstractmethod
    async def apply(self, *args: Any, **kwargs: Any) -> bool:
        """Apply the condition asynchronously."""


class Signal(Event):
    """a triggerable signal."""

    @abstractmethod
    async def trigger(self, *args: Any, **kwargs: Any) -> Any:
        """Trigger the signal asynchronously."""


class Action(Event):
    """An invokable action"""

    @abstractmethod
    async def invoke(self, *args: Any, **kwargs: Any) -> Any:
        """Invoke the action asynchronously."""


__all__ = ["Event", "Condition", "Signal", "Action"]
